@@ -1,13 +1,17 @@
 import webbrowser
-from typing import List, Literal, Optional, Union
+from typing import Any, Dict, List, Literal, Optional, Union
 
 import pandas as pd
 from pydantic import BaseModel, Extra, PositiveInt
 
 from .base import RETURN_FORMAT, FredBase, Json, JsonOrPandas, ReturnFormat
+from .utils import _convert_to_pandas
 
 
-class SeriesArgs(BaseModel):
+class SeriesApiParameters(BaseModel):
+    """Represents the parameters accepted by the FRED Series endpoints."""
+
+    series_id: Optional[str] = None
     realtime_start: Optional[str] = None
     realtime_end: Optional[str] = None
     limit: Optional[int] = None
@@ -56,7 +60,9 @@ class SeriesArgs(BaseModel):
         extra = Extra.forbid
 
 
-class SeriesSearchArgs(BaseModel):
+class SeriesSearchParameters(BaseModel):
+    search_text: Optional[str] = None
+    search_type: Optional[Literal["full_text", "series_id"]] = None
     realtime_start: Optional[str] = None
     realtime_end: Optional[str] = None
     limit: Optional[int] = None
@@ -78,10 +84,11 @@ class SeriesSearchArgs(BaseModel):
             "group_popularity",
         ]
     ] = None
+    sort_order: Optional[Literal["asc", "desc"]] = None
     filter_variable: Optional[
         Literal["frequency", "units", "seasonal_adjustment"]
     ] = None
-    filter_values: Optional[str] = None
+    filter_value: Optional[str] = None
     tag_names: Optional[str] = None
     exclude_tag_names: Optional[str] = None
 
@@ -106,13 +113,26 @@ class SeriesInfo(BaseModel):
     seasonal_adjustment_short: str
     last_updated: str
     popularity: int
-    notes: str
+    notes: Optional[str] = None
     _base_url: str = "https://fred.stlouisfed.org/series"
 
     def open_url(self):
         """Open the FRED webpage for the given series."""
         webbrowser.open(f"{self._base_url}/{self.id}", new=2)
 
+    class Config:
+        extra = Extra.allow
+
+
+class SeriesData(BaseModel):
+    """Represents the response from series/observations endpoints."""
+
+    info: SeriesInfo
+    data: Union[List[Dict[str, Any]], pd.DataFrame]
+
+    class Config:
+        arbitrary_types_allowed = True
+
 
 class FredSeries(FredBase):
     """FRED API series endpoints."""
@@ -128,20 +148,14 @@ class FredSeries(FredBase):
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/ endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/ endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
-        SeriesInfo object.
+        An instance of SeriesInfo.
         """
-        params = SeriesArgs(**kwargs)
-        response = self._get(
-            endpoint="series",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
-        )
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
+        response = self._get(endpoint="series", params=params.dict(exclude_none=True))
         return SeriesInfo(**response["seriess"][0])
 
     def get_series_categories(self, series_id: str, **kwargs) -> Json:
@@ -152,27 +166,24 @@ class FredSeries(FredBase):
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/categories endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/categories endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
         Dictionary representing the json response.
         """
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         return self._get(
             endpoint="series/categories",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
 
-    def get_series_data(
+    def get_series(
         self,
         series_id: str,
         return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
         **kwargs,
-    ) -> JsonOrPandas:
+    ) -> SeriesData:
         """Get the observations or data values for an economic data series by ID. https://fred.stlouisfed.org/docs/api/fred/series_observations.html.
 
         Parameters
@@ -182,26 +193,26 @@ class FredSeries(FredBase):
         return_format : Literal["json", "pandas"] | ReturnFormat, optional
             Define how to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/observations endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/observations endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
-        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        A SeriesData object that holds the data and metadata for the requested series id.
         """
         return_format = ReturnFormat(return_format)
 
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         response = self._get(
             endpoint="series/observations",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
 
+        series_info = self.get_series_info(series_id=series_id)
+
         if return_format == ReturnFormat.pandas:
-            return pd.DataFrame.from_dict(response["observations"])
-        return response
+            pdf = _convert_to_pandas(response["observations"])
+            return SeriesData(info=series_info, data=pdf)
+        return SeriesData(info=series_info, data=response["observations"])
 
     def get_series_releases(self, series_id: str, **kwargs) -> Json:
         """Get the FRED release for an economic data series by ID. https://fred.stlouisfed.org/docs/api/fred/series_release.html.
@@ -211,19 +222,16 @@ class FredSeries(FredBase):
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/releases endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/releases endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
         Dictionary representing the json response.
         """
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         return self._get(
-            endpoint="series/releases",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            endpoint="series/release",
+            params=params.dict(exclude_none=True),
         )
 
     def get_series_tags(self, series_id: str, **kwargs) -> Json:
@@ -234,19 +242,16 @@ class FredSeries(FredBase):
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/tags endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/tags endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
         Dictionary representing the json response.
         """
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         return self._get(
             endpoint="series/tags",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
 
     def get_series_updates(self, series_id: str, **kwargs) -> Json:
@@ -257,44 +262,38 @@ class FredSeries(FredBase):
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/updates endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/updates endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
         Dictionary representing the json response.
         """
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         return self._get(
             endpoint="series/updates",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
 
     def get_series_vintagedates(self, series_id: str, **kwargs) -> List[str]:
         """Get the dates in history when a series' data values were revised or new data values were released.
 
-        Vintage dates are the release dates for a series excluding release dates when the data for the series did not change. https://fred.stlouisfed.org/docs/api/fred/series_vintagesdates.html.
+        Vintage dates are the release dates for a series excluding release dates when the data did not change. https://fred.stlouisfed.org/docs/api/fred/series_vintagesdates.html.
 
         Parameters
         ----------
         series_id : str
             Series id of interest.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/vintagedates endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/vintagedates endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
         List of strings representing the the available vintage dates
         """
-        params = SeriesArgs(**kwargs)
+        params = SeriesApiParameters(series_id=series_id, **kwargs)
         response = self._get(
             endpoint="series/vintagedates",
-            params={
-                "series_id": series_id,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
         return response["vintage_dates"]
 
@@ -303,7 +302,7 @@ class FredSeries(FredBase):
         series_id: str,
         return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
         **kwargs,
-    ) -> JsonOrPandas:
+    ) -> SeriesData:
         """Get the observations or data values for all releases an economic data series by ID.
 
         Parameters
@@ -313,31 +312,36 @@ class FredSeries(FredBase):
         return_format : Literal["json", "pandas"] | ReturnFormat, optional
             In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/observation endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
-        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        A SeriesData object that holds the data and metadata for the requested series id.
         """
         return_format = ReturnFormat(return_format)
 
-        params = SeriesArgs(**kwargs)
-        return self.get_series_data(
+        # sanitize the kwargs to ensure the user did not
+        # in inadvertently supply values for realtime_start & realtime_end
+        _ = kwargs.pop("realtime_start", None)
+        _ = kwargs.pop("realtime_end", None)
+
+        params = SeriesApiParameters(
+            realtime_start=self._earliest_realtime_start,
+            realtime_end=self._latest_realtime_end,
+            **kwargs,
+        )
+        return self.get_series(
             series_id=series_id,
             return_format=return_format,
-            **{
-                "realtime_start": self._earliest_realtime_start,
-                "realtime_end": self._latest_realtime_end,
-                **params.dict(exclude_none=True),
-            },
+            **params.dict(exclude_none=True),
         )
 
     def get_series_initial_release(
         self,
         series_id: str,
-        return_format: Union[RETURN_FORMAT, ReturnFormat],
+        return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
         **kwargs,
-    ) -> JsonOrPandas:
+    ) -> SeriesData:
         """Get the observations or data values for the initial release of an economic data series.
 
         Includes only the the initial release of the series and excludes all revisions.
@@ -349,28 +353,26 @@ class FredSeries(FredBase):
         return_format : Literal["json", "pandas"] | ReturnFormat, optional
             In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/observation endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
-        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        A SeriesData object that holds the data and metadata for the requested series id.
         """
         return_format = ReturnFormat(return_format)
 
         # sanitize the kwargs to ensure the user did not
         # in inadvertently supply values for realtime_start & output_type
-        _ = kwargs.pop("realtime_start")
-        _ = kwargs.pop("output_type")
+        _ = kwargs.pop("realtime_start", None)
+        _ = kwargs.pop("output_type", None)
 
-        params = SeriesArgs(**kwargs)
-        return self.get_series_data(
+        params = SeriesApiParameters(
+            realtime_start=self._earliest_realtime_start, output_type=4, **kwargs
+        )
+        return self.get_series(
             series_id=series_id,
             return_format=return_format,
-            **{
-                "realtime_start": self._earliest_realtime_start,
-                "output_type": 4,
-                **params.dict(exclude_none=True),
-            },
+            **params.dict(exclude_none=True),
         )
 
     def get_series_asof_date(
@@ -379,7 +381,7 @@ class FredSeries(FredBase):
         date: str,
         return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
         **kwargs,
-    ) -> JsonOrPandas:
+    ) -> SeriesData:
         """Get the observations or data values for an economic data series made on or before a specific date.
 
         Retrieves the latest data known for the series as of the date provided. This includes any revisions to
@@ -389,28 +391,31 @@ class FredSeries(FredBase):
         ----------
         series_id : str
             Series id of interest.
-        date: str
+        date : str
             Include only data revisions made on or before this date.
-        return_format : : Literal["json", "pandas"] | ReturnFormat, optional
+        return_format : Literal["json", "pandas"] | ReturnFormat, optional
             In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/observation endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
-        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        A SeriesData object that holds the data and metadata for the requested series id.
         """
         return_format = ReturnFormat(return_format)
 
-        params = SeriesArgs(**kwargs)
-        return self.get_series_data(
+        # sanitize the kwargs to ensure the user did not
+        # in inadvertently supply values for realtime_start & realtime_end
+        _ = kwargs.pop("realtime_start", None)
+        _ = kwargs.pop("realtime_end", None)
+
+        params = SeriesApiParameters(
+            realtime_start=self._earliest_realtime_start, realtime_end=date, **kwargs
+        )
+        return self.get_series(
             series_id=series_id,
             return_format=return_format,
-            **{
-                "realtime_start": self._earliest_realtime_start,
-                "realtime_end": date,
-                **params.dict(exclude_none=True),
-            },
+            **params.dict(exclude_none=True),
         )
 
     def search_series(
@@ -425,14 +430,14 @@ class FredSeries(FredBase):
         Parameters
         ----------
         search_text : str
-            The word to match against.
-        search_type : Literal[str]
+            The text to match against.
+        search_type : Literal["full_text", "series_id"]
             Defines which type of search to preform. One of the following strings: 'full_text', 'series_id'.
-            `Parameter docs` <https://fred.stlouisfed.org/docs/api/fred/series_search.html#search_type>`_.
+            Parameter docs: https://fred.stlouisfed.org/docs/api/fred/series_search.html#search_type.
         return_format : : Literal["json", "pandas"] | ReturnFormat, optional
             In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
         **kwargs : dict, optional
-            Extra arguments to FRED API series/observation endpoint. Refer the FRED documentation for a list of all possible arguments.
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
 
         Returns
         -------
@@ -440,22 +445,92 @@ class FredSeries(FredBase):
         """
         return_format = ReturnFormat(return_format)
 
-        params = SeriesSearchArgs(**kwargs)
+        params = SeriesSearchParameters(
+            search_text=search_text,
+            search_type=search_type,
+            **kwargs,
+        )
         response = self._get(
             endpoint="series/search",
-            params={
-                "search_text": search_text,
-                "search_type": search_type,
-                **params.dict(exclude_none=True),
-            },
+            params=params.dict(exclude_none=True),
         )
 
         if return_format == ReturnFormat.pandas:
             return pd.DataFrame.from_dict(response["seriess"])
         return response
 
-    def search_series_tags(self):
-        raise NotImplementedError
+    def search_series_tags(
+        self,
+        search_text: str,
+        return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
+        **kwargs,
+    ) -> JsonOrPandas:
+        """Get the FRED tags for a series search. https://fred.stlouisfed.org/docs/api/fred/release_related_tags.html.
 
-    def search_series_related_tags(self):
-        raise NotImplementedError
+        Parameters
+        ----------
+        search_text : str
+            The text to match against.
+        return_format : : Literal["json", "pandas"] | ReturnFormat, optional
+            In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
+        **kwargs : dict, optional
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
+
+        Returns
+        -------
+        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        """
+        return_format = ReturnFormat(return_format)
+
+        params = SeriesSearchParameters(**kwargs)
+        response = self._get(
+            endpoint="series/search/tags",
+            params={
+                "series_search_text": search_text,
+                **params.dict(exclude_none=True),
+            },
+        )
+
+        if return_format == ReturnFormat.pandas:
+            return pd.DataFrame.from_dict(response["tags"])
+        return response
+
+    def search_series_related_tags(
+        self,
+        search_text: str,
+        tag_names: str,
+        return_format: Union[RETURN_FORMAT, ReturnFormat] = "pandas",
+        **kwargs,
+    ) -> JsonOrPandas:
+        """Get the related FRED tags for one or more FRED tags matching a series search. https://fred.stlouisfed.org/docs/api/fred/series_search_related_tags.html.
+
+        Parameters
+        ----------
+        search_text : str
+            The text to match against.
+        tag_names : str
+            A semicolon delimited list of tag names that series match all of.
+        return_format : : Literal["json", "pandas"] | ReturnFormat, optional
+            In what format to return the response. Must be either 'json' or 'pandas'. Defaults to 'pandas'.
+        **kwargs : dict, optional
+            Additional parameters to FRED API series/observation endpoint. Refer to the FRED documentation for a list of all possible parameters.
+
+        Returns
+        -------
+        Either a pandas dataframe or json. Defaults to pandas dataframe.
+        """
+        return_format = ReturnFormat(return_format)
+
+        params = SeriesSearchParameters(**kwargs)
+        response = self._get(
+            endpoint="series/search/related_tags",
+            params={
+                "series_search_text": search_text,
+                "tag_names": tag_names,
+                **params.dict(exclude_none=True),
+            },
+        )
+
+        if return_format == ReturnFormat.pandas:
+            return pd.DataFrame.from_dict(response["tags"])
+        return response
